@@ -17,7 +17,7 @@
  * 
  *   Time-stamp: <Tuesday 14 June 05 0:08:56 tews@debian>
  * 
- *   $Id: paqo_o.ml,v 1.12 2005/09/22 22:22:07 tews Exp $
+ *   $Id: paqo_o.ml,v 1.16 2006/07/19 22:25:06 tews Exp $
  * 
  *)
 
@@ -25,7 +25,7 @@
 open Stdpp
 open Pcaml
 
-let ocamlp4_version = "0.2";;
+let ocamlp4_version = "0.3";;
 
 Pcaml.syntax_name := "OCaml";;
 Pcaml.no_constructors_arity := false;;
@@ -71,20 +71,29 @@ let string_of_option = function
   | Some _ -> "Some _"
   | None -> "None"
 
-let mkumin loc f arg =
+let mkexprident _loc ids = match ids with
+  | [] -> Stdpp.raise_with_loc _loc (Stream.Error "illegal long identifier")
+  | id :: ids ->
+      let rec loop m = function
+        | id :: ids -> loop <:expr< $m$ . $id$ >> ids
+        | [] -> m
+      in
+	loop id ids
+
+let mkumin _loc f arg =
   match (f, arg) with
     | ("-", <:expr< $int:n$ >>) when int_of_string n > 0 ->
 	<:expr< $int:"-" ^ n$ >>
     | ("-", <:expr< $int32:n$ >>) when (Int32.of_string n) > 0l ->
-	let loc = if !pa_o_compatibility then MLast.loc_of_expr arg else loc
+	let _loc = if !pa_o_compatibility then MLast.loc_of_expr arg else _loc
 	in
 	  <:expr< $int32:"-" ^ n$ >>
     | ("-", <:expr< $int64:n$ >>) when (Int64.of_string n) > 0L ->
-	let loc = if !pa_o_compatibility then MLast.loc_of_expr arg else loc
+	let _loc = if !pa_o_compatibility then MLast.loc_of_expr arg else _loc
 	in
 	  <:expr< $int64:"-" ^ n$ >>
     | ("-", <:expr< $nativeint:n$ >>) when (Nativeint.of_string n) > 0n ->
-	let loc = if !pa_o_compatibility then MLast.loc_of_expr arg else loc
+	let _loc = if !pa_o_compatibility then MLast.loc_of_expr arg else _loc
 	in
 	  <:expr< $nativeint:"-" ^ n$ >>
     | (_, <:expr< $flo:n$ >>) when float_of_string n > 0.0 ->
@@ -96,11 +105,11 @@ let mkumin loc f arg =
     (* there is no tail in immediate list in concreate syntax
      * -> hardcode this, erase the "last" argument
      *)
-let mklistexp loc =
+let mklistexp _loc =
   let rec loop top = function
     | [] -> <:expr< [] >> 
     | e1 :: el ->
-        let loc = if top then loc else (fst (MLast.loc_of_expr e1), snd loc) 
+        let _loc = if top then _loc else (fst (MLast.loc_of_expr e1), snd _loc) 
 	in
           <:expr< $e1$ :: $loop false el$ >> 
   in
@@ -108,11 +117,11 @@ let mklistexp loc =
 
 
     (* same as mklistexp: erase the "last" argument *)
-let mklistpat loc =
+let mklistpat _loc =
   let rec loop top = function
     | [] -> <:patt< [] >>
     | p1 :: pl ->
-        let loc = if top then loc else (fst (MLast.loc_of_patt p1), snd loc) 
+        let _loc = if top then _loc else (fst (MLast.loc_of_patt p1), snd _loc) 
 	in
           <:patt< $p1$ :: $loop false pl$ >>
   in
@@ -314,17 +323,17 @@ let rec is_expr_constr_call = function
   | _ -> false 
 
 
-let rec constr_expr_arity loc = function
+let rec constr_expr_arity _loc = function
   | <:expr< $uid:c$ >> ->
     (try 
        List.assoc c !constr_arity
      with 
 	 Not_found -> 0 
     )
-  | <:expr< $uid:_$.$e$ >> -> constr_expr_arity loc e
+  | <:expr< $uid:_$.$e$ >> -> constr_expr_arity _loc e
   | <:expr< $e$ $_$ >> ->
       if is_expr_constr_call e then
-        Stdpp.raise_with_loc loc (Stream.Error "currified constructor")
+        Stdpp.raise_with_loc _loc (Stream.Error "currified constructor")
       else 1
   | _ -> 1
 
@@ -335,16 +344,16 @@ let rec is_patt_constr_call = function
   | _ -> false 
 
 
-let rec constr_patt_arity loc = function
+let rec constr_patt_arity _loc = function
   | <:patt< $uid:c$ >> ->
       (try List.assoc c !constr_arity
        with 
 	 | Not_found -> 0 
       )
-  | <:patt< $uid:_$.$p$ >> -> constr_patt_arity loc p
+  | <:patt< $uid:_$.$p$ >> -> constr_patt_arity _loc p
   | <:patt< $p$ $_$ >> ->
       if is_patt_constr_call p then
-        Stdpp.raise_with_loc loc (Stream.Error "currified constructor")
+        Stdpp.raise_with_loc _loc (Stream.Error "currified constructor")
       else 1
   | _ -> 1 
 
@@ -389,7 +398,7 @@ let rec patt_lid = function
  * However, pa_o accepts the above expression, so we do not divert 
  * from the diversion.
  *)
-let bigarray_get loc arr arg =
+let bigarray_get _loc arr arg =
   let coords = match arg with
     | <:expr< $tup:el$ >> -> el
     | _ -> [arg]
@@ -401,7 +410,7 @@ let bigarray_get loc arr arg =
       | coords -> <:expr< $arr$. { $tup:coords$ } >>
 
 
-let bigarray_set loc var newval =
+let bigarray_set _loc var newval =
   match var with
     | <:expr< $arr$.{ $c1$ } >> ->
       Some <:expr< $arr$.{ $c1$ } <- $newval$ >>
@@ -581,6 +590,8 @@ EXTEND
           <:expr< for $i$ = $e1$ $to:df$ $e2$ do $seq:get_seq e$ done >>
       | "while"; e1 = SELF; "do"; e2 = SELF; "done" ->
           <:expr< while $e1$ do $seq:get_seq e2$ done >>
+      | "object"; cspo = OPT class_self_patt; cf = class_structure; "end" ->
+          <:expr< object $opt:cspo$ $list:cf$ end >>
       ]
     | [ e = SELF; ","; el = LIST1 NEXT SEP "," ->
           <:expr< $tup: e :: el$ >> ]
@@ -589,7 +600,7 @@ EXTEND
       [ e1 = SELF; ":="; e2 = expr LEVEL "expr1" ->
           <:expr< $e1$ := $e2$ >>
       | e1 = SELF; "<-"; e2 = expr LEVEL "expr1" ->
-          (match bigarray_set loc e1 e2 with
+          (match bigarray_set _loc e1 e2 with
              | Some e -> e
              | None -> <:expr< $e1$ <- $e2$ >> 
 	  ) ]
@@ -637,12 +648,12 @@ EXTEND
       | e1 = SELF; "lsr"; e2 = SELF -> <:expr< $e1$ lsr $e2$ >>
       | e1 = SELF; op = infixop4; e2 = SELF -> <:expr< $lid:op$ $e1$ $e2$ >> ]
     | "unary minus" NONA
-      [ "-"; e = SELF -> <:expr< $mkumin loc "-" e$ >>
-      | "-."; e = SELF -> <:expr< $mkumin loc "-." e$ >> ]
+      [ "-"; e = SELF -> <:expr< $mkumin _loc "-" e$ >>
+      | "-."; e = SELF -> <:expr< $mkumin _loc "-." e$ >> ]
 
     | "apply" LEFTA
 	[ e1 = SELF; e2 = SELF ->
-            (match constr_expr_arity loc e1 with
+            (match constr_expr_arity _loc e1 with
 	       | 1 -> <:expr< $e1$ $e2$ >>
                | _ ->
 		   (match e2 with
@@ -658,7 +669,7 @@ EXTEND
     | "." LEFTA
       [ e1 = SELF; "."; "("; e2 = SELF; ")" -> <:expr< $e1$ .( $e2$ ) >>
       | e1 = SELF; "."; "["; e2 = SELF; "]" -> <:expr< $e1$ .[ $e2$ ] >>
-      | e1 = SELF; "."; "{"; e2 = SELF; "}" -> bigarray_get loc e1 e2
+      | e1 = SELF; "."; "{"; e2 = SELF; "}" -> bigarray_get _loc e1 e2
       | e1 = SELF; "."; e2 = SELF -> <:expr< $e1$ . $e2$ >> ]
     | "~-" NONA
       [ "!"; e = SELF -> <:expr< ! $e$ >>
@@ -676,11 +687,11 @@ EXTEND
       | c = CHAR -> <:expr< $chr:c$ >>
       | UIDENT "True" -> <:expr< $uid:" True"$ >>
       | UIDENT "False" -> <:expr< $uid:" False"$ >>
-      | i = expr_ident -> i
+      | ids = expr_ident -> mkexprident _loc ids
       | s = "false" -> <:expr< false >>
       | s = "true" -> <:expr< true >>
       | "["; "]" -> <:expr< [] >>
-      | "["; el = expr1_semi_list; "]" -> <:expr< $mklistexp loc el$ >>
+      | "["; el = expr1_semi_list; "]" -> <:expr< $mklistexp _loc el$ >>
       | "[|"; "|]" -> <:expr< [| |] >>
       | "[|"; el = expr1_semi_list; "|]" -> <:expr< [| $list:el$ |] >>
       | "{"; test_label_eq; lel = lbl_expr_list; "}" ->
@@ -705,7 +716,7 @@ EXTEND
             with
 	      | Not_found | Failure _ -> (Token.nowhere, x)
           in
-            Pcaml.handle_expr_locate loc x
+            Pcaml.handle_expr_locate _loc x
       | x = QUOTATION ->
           let x =
             try
@@ -716,14 +727,14 @@ EXTEND
             with
 	      | Not_found -> ("", x) 
           in
-            Pcaml.handle_expr_quotation loc x ] ]
+            Pcaml.handle_expr_quotation _loc x ] ]
   ;
 
 
   let_binding:
     [ [ p = patt; e = fun_binding ->
           match patt_lid p with
-            | Some (loc, i, pl) ->
+            | Some (_loc, i, pl) ->
 		let e =
                   List.fold_left (fun e p -> <:expr< fun $p$ -> $e$ >>) e pl
 		in
@@ -762,16 +773,13 @@ EXTEND
   ;
   expr_ident:
     [ RIGHTA
-      [ i = LIDENT -> <:expr< $lid:i$ >>
-      | i = UIDENT -> <:expr< $uid:i$ >>
-      | i = UIDENT; "."; j = SELF ->
-          let rec loop m = function
-            | <:expr< $x$ . $y$ >> -> loop <:expr< $m$ . $x$ >> y
-            | e -> <:expr< $m$ . $e$ >> 
-          in
-            loop <:expr< $uid:i$ >> j
+      [ i = LIDENT -> [ <:expr< $lid:i$ >> ]
+      | i = UIDENT -> [ <:expr< $uid:i$ >> ]
       | i = UIDENT; "."; "("; j = operator_rparen ->
-          <:expr< $uid:i$ . $lid:j$ >> ] ]
+         [ <:expr< $uid:i$ >> ; <:expr< $lid:j$ >> ]
+      | i = UIDENT; "."; j = SELF ->  <:expr< $uid:i$ >> :: j
+      ]
+    ]
   ;
 
 
@@ -794,7 +802,7 @@ EXTEND
       [ p1 = SELF; "::"; p2 = SELF -> <:patt< $p1$ :: $p2$ >> ]
     | LEFTA
       [ p1 = SELF; p2 = SELF ->
-          (match constr_patt_arity loc p1 with
+          (match constr_patt_arity _loc p1 with
              | 1 -> <:patt< $p1$ $p2$ >>
              | n ->
 		 let p2 = 
@@ -837,7 +845,7 @@ EXTEND
       | s = "false" -> <:patt< false >>
       | s = "true" -> <:patt< true >>
       | "["; "]" -> <:patt< [] >>
-      | "["; pl = patt_semi_list; "]" -> <:patt< $mklistpat loc pl$ >>
+      | "["; pl = patt_semi_list; "]" -> <:patt< $mklistpat _loc pl$ >>
       | "[|"; "|]" -> <:patt< [| |] >>
       | "[|"; pl = patt_semi_list; "|]" -> <:patt< [| $list:pl$ |] >>
 
@@ -883,64 +891,7 @@ EXTEND
   type_declaration:
     [ [ tpl = type_parameters; n = type_patt; "="; tk = type_kind;
         cl = LIST0 constrain ->
-          (* (n, tpl, tk, cl) *)
-	  let (nn, ntpl, ntk, ncl) as td = match (fst tk) with
-	    | (true, None, Some ldl, None)    ->
-	     	<:typedef< ( $list:tpl$ ) $locstr:n$ = 
-	                     private { $list:ldl$ } $list:cl$ >>
-	     
-	    | (false, None, Some ldl, None)   ->
-		<:typedef< ( $list:tpl$ ) $locstr:n$ = 
-	                                    { $list:ldl$ } $list:cl$ >>
-	     
-	    | (true, None, None, Some cdl)    ->
-	     	<:typedef< ( $list:tpl$ ) $locstr:n$ = 
-                              private $list:cdl$ $list:cl$ >>
-	     
-	    | (false, None, None, Some cdl)   ->
-	     	<:typedef< ( $list:tpl$ ) $locstr:n$ = 
-	                      $list:cdl$ $list:cl$ >>
-	     
-            | (false, Some t, None, None)     ->
-	     	<:typedef< ( $list:tpl$ ) $locstr:n$ = $t$ $list:cl$ >>
-	     
-	    | (true, Some t, Some ldl, None)  ->
-	     	<:typedef< ( $list:tpl$ ) $locstr:n$ = $t$ = 
-	                      private { $list:ldl$ } $list:cl$ >> 
-
-	    | (false, Some t, Some ldl, None) ->
-	     	<:typedef< ( $list:tpl$ ) $locstr:n$ = $t$ = 
-	                      { $list:ldl$ } $list:cl$ >>
-
-	    | (true, Some t, None, Some cdl)  ->
-	     	<:typedef< ( $list:tpl$ ) $locstr:n$ = $t$ = 
-                              private $list:cdl$ $list:cl$ >>
-	     
-	    | (false, Some t, None, Some cdl) ->
-		<:typedef< ( $list:tpl$ ) $locstr:n$ = 
-			      $t$ = $list:cdl$ $list:cl$ >>
-
-	    | (b, o1, o2, o3) -> 
-		Printf.fprintf stderr 
-		  "assertion in type_declaration: (%B, %s, %s, %s)\n"
-		  b (string_of_option o1) 
-		  (string_of_option o2) (string_of_option o3);
-		assert false
-	  in
-	    if !pa_o_compatibility then
-	      (match (fst tk) with
-		 | (false, Some _, None, None) -> ()
-		 | _ ->
-		     begin
-		       Obj.set_field (Obj.repr ntk) 0 (Obj.repr (snd tk));
-		       (match ntk with
-			  | MLast.TyMan(_,_,otk) -> 
-			      Obj.set_field (Obj.repr otk) 0 (Obj.repr (snd tk))
-			  | _ -> ()
-		       );
-		     end
-	      );
-	    td
+	  <:typedef< ( $list:tpl$ ) $locstr:n$ = $tk$ $list:cl$ >>
 
       |	tpl = type_parameters; n = type_patt; cl = LIST0 constrain ->
           (* (n, tpl, <:ctyp< '$choose_tvar tpl$ >>, cl)  *)
@@ -949,44 +900,25 @@ EXTEND
       ] ]
   ;
   type_patt:
-    [ [ n = LIDENT -> (loc, n) ] ]
+    [ [ n = LIDENT -> (_loc, n) ] ]
   ;
   constrain:
     [ [ "constraint"; t1 = ctyp; "="; t2 = ctyp -> (t1, t2) ] ]
   ;
   type_kind:
-    [ [ "private"; "{"; ldl = label_declarations; "}" ->
-          (* <:ctyp< private { $list:ldl$ } >> *)
-      	  ((true, None, Some ldl, None), loc)
-      | "private"; OPT "|"; 
-        cdl = LIST1 constructor_declaration SEP "|" -> 
-       	  (* <:ctyp< private [ $list:cdl$ ] >> *)
-       	  ((true, None, None, Some cdl), loc)      
+    [ [ "private"; tk = type_kind -> <:type_info< private $tk$ >>
       | test_constr_decl; OPT "|";
         cdl = LIST1 constructor_declaration SEP "|" -> 
-	  (* <:ctyp< [ $list:cdl$ ] >> *)
-	  ((false, None, None, Some cdl), loc)
-      | t = ctyp -> 
-      	  (* <:ctyp< $t$ >> *)
-      	  ((false, Some t, None, None), loc)
-      | t = ctyp; "="; "private"; "{"; ldl = label_declarations; "}" ->
-          (* <:ctyp< $t$ == private { $list:ldl$ } >> *)
-      	  ((true, Some t, Some ldl, None), loc)
+	  <:type_info< $list:cdl$ >>
+      | t = ctyp -> <:type_info< $t$ >>
+      | t = ctyp; "="; "private"; tk = type_kind ->
+          <:type_info< $t$ = private $tk$ >>
       | t = ctyp; "="; "{"; ldl = label_declarations; "}" ->
-          (* <:ctyp< $t$ == { $list:ldl$ } >> *)
-      	  ((false, Some t, Some ldl, None), loc)
-      | t = ctyp; "="; "private"; OPT "|"; 
-      	  cdl = LIST1 constructor_declaration SEP "|" ->
-          (* <:ctyp< $t$ == private [ $list:cdl$ ] >> *)
-      	    ((true, Some t, None, Some cdl), loc)
-      
+          <:type_info< $t$ = { $list:ldl$ } >>
       | t = ctyp; "="; OPT "|"; cdl = LIST1 constructor_declaration SEP "|" ->
-          (* <:ctyp< $t$ == [ $list:cdl$ ] >> *)
-      	  ((false, Some t, None, Some cdl), loc)
+          <:type_info< $t$ = $list:cdl$ >>
       | "{"; ldl = label_declarations; "}" ->
-          (* <:ctyp< { $list:ldl$ } >>  *)
-      	  ((false, None, Some ldl, None), loc)
-      ] ]
+          <:type_info< { $list:ldl$ } >> ] ]
   ;
   type_parameters:
     [ [ -> (* empty *) []
@@ -1000,8 +932,8 @@ EXTEND
   ;
   constructor_declaration:
     [ [ ci = UIDENT; "of"; cal = LIST1 ctyp LEVEL "ctyp1" SEP "*" ->
-          (loc, ci, cal)
-      | ci = UIDENT -> (loc, ci, []) ] ]
+          (_loc, ci, cal)
+      | ci = UIDENT -> (_loc, ci, []) ] ]
   ;
   label_declarations:
     [ [ ld = label_declaration; ";"; ldl = SELF -> ld :: ldl
@@ -1009,8 +941,8 @@ EXTEND
       | ld = label_declaration -> [ld] ] ]
   ;
   label_declaration:
-    [ [ i = LIDENT; ":"; t = poly_type -> (loc, i, false, t)
-      | "mutable"; i = LIDENT; ":"; t = poly_type -> (loc, i, true, t) ] ]
+    [ [ i = LIDENT; ":"; t = poly_type -> (_loc, i, false, t)
+      | "mutable"; i = LIDENT; ":"; t = poly_type -> (_loc, i, true, t) ] ]
   ;
 
 
@@ -1101,7 +1033,7 @@ EXTEND
   class_declaration:
     [ [ vf = OPT "virtual"; ctp = class_type_parameters; i = LIDENT;
         cfb = class_fun_binding ->
-          {MLast.ciLoc = loc; MLast.ciVir = o2b vf; MLast.ciPrm = ctp;
+          {MLast.ciLoc = _loc; MLast.ciVir = o2b vf; MLast.ciPrm = ctp;
            MLast.ciNam = i; MLast.ciExp = cfb} ] ]
   ;
   class_fun_binding:
@@ -1112,8 +1044,8 @@ EXTEND
           <:class_expr< fun $p$ -> $cfb$ >> ] ]
   ;
   class_type_parameters:
-    [ [ -> (loc, [])
-      | "["; tpl = LIST1 type_parameter SEP ","; "]" -> (loc, tpl) ] ]
+    [ [ -> (_loc, [])
+      | "["; tpl = LIST1 type_parameter SEP ","; "]" -> (_loc, tpl) ] ]
   ;
   class_fun_def:
     [ [ p = patt LEVEL "simple"; "->"; ce = class_expr ->
@@ -1235,13 +1167,13 @@ EXTEND
   class_description:
     [ [ vf = OPT "virtual"; ctp = class_type_parameters; n = LIDENT; ":";
         ct = class_type ->
-          {MLast.ciLoc = loc; MLast.ciVir = o2b vf; MLast.ciPrm = ctp;
+          {MLast.ciLoc = _loc; MLast.ciVir = o2b vf; MLast.ciPrm = ctp;
            MLast.ciNam = n; MLast.ciExp = ct} ] ]
   ;
   class_type_declaration:
     [ [ vf = OPT "virtual"; ctp = class_type_parameters; n = LIDENT; "=";
         cs = class_signature ->
-          {MLast.ciLoc = loc; MLast.ciVir = o2b vf; MLast.ciPrm = ctp;
+          {MLast.ciLoc = _loc; MLast.ciVir = o2b vf; MLast.ciPrm = ctp;
            MLast.ciNam = n; MLast.ciExp = cs} ] ]
   ;
 
@@ -1350,7 +1282,7 @@ EXTEND
       | EOI -> ([], false) ] ]
   ;
   str_item_semi:
-    [ [ si = str_item; OPT ";;" -> (si, loc) ] ]
+    [ [ si = str_item; OPT ";;" -> (si, _loc) ] ]
   ;
 END;
 ;;
