@@ -17,7 +17,7 @@
  * 
  *   Time-stamp: <Tuesday 14 June 05 0:08:56 tews@debian>
  * 
- *   $Id: pp_q.ml,v 1.12 2005/10/06 15:34:08 tews Exp $
+ *   $Id: pp_q.ml,v 1.15 2005/11/14 22:46:41 tews Exp $
  * 
  *)
 
@@ -39,6 +39,14 @@ let rec pp_id_list pp ids =
 let rec pp_ctyp pp ast = 
   let astloc = MLast.loc_of_ctyp ast in
   let pl mesg = print_location pp mesg astloc in
+  let pl_rec_1 s t =
+      begin
+	pl s;
+	pp_ctyp pp t;
+	if !terminate_structures then
+	  IB.endline pp ("end " ^ s)
+      end
+  in
   let pl_rec_2 s t1 t2 =
       begin
 	pl s;
@@ -51,6 +59,7 @@ let rec pp_ctyp pp ast =
   let _ = IB.indent pp in
   let _ = match ast with
     | <:ctyp< $t1$ == $t2$>> -> pl_rec_2 "type manifest" t1 t2
+    | <:ctyp< private $t1$>> -> pl_rec_1 "private type" t1
     | <:ctyp< $t1$ as $t2$>> -> pl_rec_2 "type alias" t1 t2
 
     | <:ctyp< ! $list:tvl$ . $t$>> ->
@@ -86,22 +95,6 @@ let rec pp_ctyp pp ast =
 	(fun ctyp -> pp_ctyp pp ctyp)
 	types
 
-    | <:ctyp< private [ $list:cdl$ ] >> -> 
-	 begin
-	   pl (Printf.sprintf "private variant type with %d variants"
-		 (List.length cdl));
-	   pp_variant_decls pp cdl;
-	   if !terminate_structures then
-	     IB.endline pp "end priv variant type"
-	 end
-    | <:ctyp< private { $list:ldl$ } >> ->
-	 begin
-	   pl (Printf.sprintf "private record type with %d fields"
-		 (List.length ldl));
-	   pp_label_decls pp ldl;
-	   if !terminate_structures then
-	     IB.endline pp "end priv record type"
-	 end
     | <:ctyp< [ $list:cdl$ ] >> -> 
 	 begin
 	   pl (Printf.sprintf "variant type with %d variants"
@@ -535,6 +528,31 @@ let rec pp_expr pp ast =
 	  IB.endline pp "end while";
       end
 
+    (* 
+     *   no quotations for immediate objects in q_MLast
+     * | <:expr< object $opt:spopt$ $list:oitems$ end >> -> 
+     *)
+    | MLast.ExObj(loc, spopt, oitems) ->
+      begin
+	pl (Printf.sprintf "immediate object with %d items" 
+	      (List.length oitems));
+	IB.indent pp;
+	(match spopt with
+	   | None -> IB.endline pp "no self pattern"
+	   | Some patt -> 
+	       begin
+		 IB.add_string pp "self pattern:\n";
+		 pp_patt pp patt;
+	       end
+	);
+	iterate_items pp None "" "%d" "item %d:"
+	  (fun citem -> pp_class_str_item pp citem)
+	  oitems;
+	IB.unindent pp;
+	if !terminate_structures then
+	  IB.endline pp "end immediate object"
+      end
+
     (* level "tuple" *)
     | <:expr< ($a$, $b$) >> -> pl_rec_2 "two tuple" a b
     | <:expr< ($list:els$) >> ->
@@ -754,14 +772,6 @@ let rec pp_expr pp ast =
 and pp_patt pp ast =
   let astloc = MLast.loc_of_patt ast in
   let pl mesg = print_location pp mesg astloc in
-  let pl_rec_1 s t =
-      begin
-	pl s;
-	pp_patt pp t;
-	if !terminate_structures then
-	  IB.endline pp ("end " ^ s)
-      end
-  in
   let pl_rec_2 s e1 e2 =
       begin
 	pl s;
