@@ -17,7 +17,7 @@
  * 
  *   Time-stamp: <Tuesday 14 June 05 0:08:56 tews@debian>
  * 
- *   $Id: pp_qo.ml,v 1.25 2005/10/06 15:34:08 tews Exp $
+ *   $Id: pp_qo.ml,v 1.27 2005/11/07 16:19:35 tews Exp $
  * 
  *)
 
@@ -281,11 +281,91 @@ let pp_variant_decls pp decls =
     )
     decls
 
-let loc_of_kind_in_decl (_,_,kind,_) = MLast.loc_of_ctyp kind
+let loc_of_inner_manifest = function
+  | MLast.TyMan(_,_,kind) -> MLast.loc_of_ctyp kind
+  | _ -> assert false
 
-let loc_of_inner_kind_in_decl = function
-  | (_,_,MLast.TyMan(_,_,kind),_) -> Some(MLast.loc_of_ctyp kind)
-  | _ -> None
+let rec pp_type_info pp ast =
+  let astloc = MLast.loc_of_ctyp ast in
+  let pl mesg = print_location pp mesg astloc in
+  let _ = IB.indent pp in
+  let _ = match ast with
+    | <:type_info< private $ti$ >> ->
+      begin
+	pl "private type";
+	pp_type_info pp ti;
+	if !terminate_structures then
+	  IB.endline pp "end private type"
+      end
+    | <:type_info< $list:cdl$ >> ->
+      begin
+	pl (Printf.sprintf "variant type with %d variants" (List.length cdl));
+	pp_variant_decls pp cdl;
+	if !terminate_structures then
+	  IB.endline pp "end variant"
+      end
+    | <:type_info< $t$ = private $ti$ >> ->
+      let snd_man_loc = loc_of_inner_manifest ast 
+      in
+	begin 
+	  pl "private manifest type";
+	  IB.indent pp;
+	  IB.endline pp "name:";
+	  pp_ctyp pp t;
+	  print_location pp "type info:" snd_man_loc;
+	  pp_type_info pp ti;
+	  IB.unindent pp;
+	  if !terminate_structures then
+	    IB.endline pp "end private manifest"
+	end
+    | <:type_info< $t$ = { $list:labels$ } >> ->
+      let snd_man_loc = loc_of_inner_manifest ast
+      in
+	begin 
+	  pl (Printf.sprintf "manifest record with %d labels" 
+		(List.length labels));
+	  IB.indent pp;
+	  IB.endline pp "name:";
+	  pp_ctyp pp t;
+	  print_location pp "fields:" snd_man_loc;
+	  pp_label_decls pp labels;
+	  IB.unindent pp;
+	  if !terminate_structures then
+	    IB.endline pp "end manifest record"
+	end
+    | <:type_info< $t$ = $list:cdl$ >> ->
+      let snd_man_loc = loc_of_inner_manifest ast
+      in
+	begin 
+	  pl (Printf.sprintf "manifest variant type with %d variants" 
+		(List.length cdl));
+	  IB.indent pp;
+	  IB.endline pp "name:";
+	  pp_ctyp pp t;
+	  print_location pp "constructors:" snd_man_loc;
+	  pp_variant_decls pp cdl;
+	  IB.unindent pp;
+	  if !terminate_structures then
+	    IB.endline pp "end manifest variant"
+	end
+    | <:type_info< { $list:labels$ } >> ->
+      begin
+	pl (Printf.sprintf "record with %d labels" (List.length labels));
+	pp_label_decls pp labels;
+	if !terminate_structures then
+	  IB.endline pp "end record"
+      end
+    | ptyp -> 
+	begin
+	  pl "type abbreviation";
+	  pp_ctyp pp ptyp;
+	  if !terminate_structures then
+	    IB.endline pp "type abbrev"
+	end
+    (* | _ -> (pl "unknown class type"; abort_maybe()) *)
+  in
+    IB.unindent pp
+  
 
 let end_of_typedef (_, _, tk, cl) =
   try
@@ -295,162 +375,66 @@ let end_of_typedef (_, _, tk, cl) =
 
 let pp_typedef pp start decl =
   let end_loc = end_of_typedef decl in
-  let kind_loc = loc_of_kind_in_decl decl in
-  let tyman_kind_loc = loc_of_inner_kind_in_decl decl in
-  let pl mesg = 
-    begin 
-      print_location pp mesg (start, end_loc);
-      print_location pp "kind loc:" kind_loc;
-      match tyman_kind_loc with
-	| None -> ()
-	| Some loc -> print_location pp "inner kind loc:" loc;
-    end
+  let pl mesg = print_location pp mesg (start, end_loc);
   in 
   let _ = IB.indent pp in
   let _ = 
     (match decl with
        | <:typedef< ( $list:tpl$ ) $locstr:lid$ = '$oid$ $list:cl$>> 
-	 when not (List.exists (fun (id,_) -> id = oid) tpl) -> 
-	 begin
-	   IB.printf pp "abstract type %s (%s)\n" 
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf 
-		 "(rhs tv %s) with %d parameters and %d constraints"
-		 oid (List.length tpl) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end abstract type"
-	 end
+       	 when not (List.exists (fun (id,_) -> id = oid) tpl) -> 
+	 let snd_loc = 
+	   match decl with
+	     | <:typedef< ( $list:_$ ) $locstr:_$ = $ot$ $list:_$>> ->
+	       MLast.loc_of_ctyp ot
+	 in
+       	   begin 
+       	     pl (Printf.sprintf 
+		   "type declaration %s (%s)\nwith %d tp and %d constraints"
+       		   (snd lid) (string_of_loc (fst lid))
+       		   (List.length tpl) (List.length cl));
+	     IB.indent pp;
+	     print_location pp "abstract_type" snd_loc;
+       	     IB.printf pp "rhs tv %s\n" oid;
+	     (match tpl with 
+		| [] -> ()
+		| _ -> 
+		    IB.endline pp "type parameters:";
+       		    pp_type_params pp tpl;
+	     );
+	     (match cl with 
+		| [] -> ()
+		| _ -> IB.endline pp "constraints:";
+       		    pp_type_constraints pp cl;
+	     );
+	     IB.unindent pp;
+       	     if !terminate_structures then
+       	       IB.endline pp "end abstract type"
+       	   end
 
-       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = 
-       	   private { $list:labels$ } $list:cl$>> -> 
-	 begin
-	   IB.printf pp "private record type %s (%s)\n"
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf "%d tp, %d fields, %d constraints"
-		 (List.length tpl) (List.length labels) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_label_decls pp labels;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end priv record"
-	 end
-	 
-       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = private $list:cdl$ $list:cl$>> -> 
-	 begin
-	   IB.printf pp "private variant type %s (%s)\n"
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf "%d tp, %d variants, %d constraints"
-		 (List.length tpl) (List.length cdl) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_variant_decls pp cdl;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end priv variant type"
-	 end
+       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = $tk$ $list:cl$>> -> 
+       	 begin
+       	   pl (Printf.sprintf 
+		 "type declaration %s (%s)\nwith %d tp and %d constraints"
+       		 (snd lid) (string_of_loc (fst lid))
+       		 (List.length tpl) (List.length cl));
+	   (match tpl with 
+	      | [] -> ()
+	      | _ -> 
+		  IB.endline pp "type parameters:";
+       		  pp_type_params pp tpl;
+	   );
+	   IB.endline pp "type information:";
+	   pp_type_info pp tk;
+	   (match cl with 
+	      | [] -> ()
+	      | _ -> IB.endline pp "constraints:";
+       		  pp_type_constraints pp cl;
+	   );
+       	   if !terminate_structures then
+       	     IB.endline pp "end type declaration"
+       	 end
 
-       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = $list:cdl$ $list:cl$>> -> 
-	 begin
-	   IB.printf pp "variant type %s (%s)\n"
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf "%d tp, %d variants, %d constraints"
-		 (List.length tpl) (List.length cdl) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_variant_decls pp cdl;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end variant type"
-	 end
-	 
-       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = $typ$ 
-	    = private { $list:labels$ } $list:cl$>> -> 
-	 begin
-	   IB.printf pp "reexport private record type %s (%s)\n"
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf "%d tp, %d fields, %d constraints"
-		 (List.length tpl) (List.length labels) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_ctyp pp typ;
-	   pp_label_decls pp labels;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end reex priv record"
-	 end
-	 
-       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = $typ$ 
-	    = { $list:labels$ } $list:cl$>> -> 
-	 begin
-	   IB.printf pp "reexport record type %s (%s)\n"
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf "%d tp, %d fields, %d constraints"
-		 (List.length tpl) (List.length labels) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_ctyp pp typ;
-	   pp_label_decls pp labels;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end reex record"
-	 end
-	 
-       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = $typ$ 
-	   = private $list:cdl$ $list:cl$>> -> 
-	 begin
-	   IB.printf pp "reexport private variant type %s (%s)\n"
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf "%d tp, %d variants, %d constraints"
-		 (List.length tpl) (List.length cdl) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_ctyp pp typ;
-	   pp_variant_decls pp cdl;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end reex priv variant"
-	 end
-
-       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = $typ$ 
-	   = $list:cdl$ $list:cl$>> -> 
-	 begin
-	   IB.printf pp "reexport variant type %s (%s)\n"
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf "%d tp, %d variants, %d constraints"
-		 (List.length tpl) (List.length cdl) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_ctyp pp typ;
-	   pp_variant_decls pp cdl;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end reex variant"
-	 end
-
-       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = 
-       	   { $list:labels$ } $list:cl$>> -> 
-	 begin
-	   IB.printf pp "record type %s (%s)\n"
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf "%d tp, %d fields, %d constraints"
-		 (List.length tpl) (List.length labels) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_label_decls pp labels;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end record"
-	 end
-
-       | <:typedef< ( $list:tpl$ ) $locstr:lid$ = $typ$ $list:cl$>> -> 
-	 begin
-	   IB.printf pp "type abbrev %s (%s)\n"
-	     (snd lid) (string_of_loc (fst lid));
-	   pl (Printf.sprintf "%d tp, %d constraints"
-		 (List.length tpl) (List.length cl));
-	   pp_type_params pp tpl;
-	   pp_ctyp pp typ;
-	   pp_type_constraints pp cl;
-	   if !terminate_structures then
-	     IB.endline pp "end type abbrev"
-	 end
-
-       (* | _ -> (pl "unknown type decl"; abort_maybe()) *)
+    (* | _ -> (pl "unknown type decl"; abort_maybe()) *)
     )
   in
     IB.unindent pp;
@@ -529,10 +513,10 @@ let rec pp_expr pp ast =
     (* level "expr1" *)
     | <:expr<let a b : $ctyp$ = 2 in 3>> when !test_more_quotations -> 
       begin
-	pl "let a b : <type follows> = 2 in 3";
-	pp_ctyp pp ctyp;
-	if !terminate_structures then
-	  IB.endline pp "end let a b"
+    	pl "let a b : <type follows> = 2 in 3";
+    	pp_ctyp pp ctyp;
+    	if !terminate_structures then
+    	  IB.endline pp "end let a b"
       end
     | <:expr<let $opt:r$ $list:bindings$ in $e$>> ->
       begin
@@ -705,6 +689,27 @@ let rec pp_expr pp ast =
 	if !terminate_structures then
 	  IB.endline pp "end while";
       end
+    | <:expr< object $opt:spopt$ $list:oitems$ end >> -> 
+      begin
+	pl (Printf.sprintf "immediate object with %d items" 
+	      (List.length oitems));
+	IB.indent pp;
+	(match spopt with
+	   | None -> IB.endline pp "no self pattern"
+	   | Some patt -> 
+	       begin
+		 IB.add_string pp "self pattern:\n";
+		 pp_patt pp patt;
+	       end
+	);
+	iterate_items pp None "" "%d" "item %d:"
+	  (fun citem -> pp_class_str_item pp citem)
+	  oitems;
+	IB.unindent pp;
+	if !terminate_structures then
+	  IB.endline pp "end immediate object"
+      end
+
 
     (* level "tuple" *)
     | <:expr< $a$, $b$ >> -> pl_rec_2 "two tuple" a b
@@ -1496,7 +1501,7 @@ and pp_class_str_item pp ast =
     | <:class_str_item< val $opt:mf$ $lid$ = $expr$ >> ->
       begin
 (* Hendrik: write object val *)
-	pl (Printf.sprintf "val %s (%s)" lid
+	pl (Printf.sprintf "object val %s (%s)" lid
 	      (if mf then "mutable" else "nonmutable"));
 	pp_expr pp expr;
       end
