@@ -15,14 +15,14 @@
  *   Library General Public License in the file LICENCE in this or one
  *   of the parent directories for more details.
  * 
- *   Time-stamp: <Friday 23 September 05 0:18:06 tews@debian>
+ *   Time-stamp: <Thursday 20 July 06 0:06:02 tews@debian>
  * 
- *   $Id: qo_MLast.ml,v 1.25 2005/09/22 22:22:07 tews Exp $
+ *   $Id: qo_MLast.ml,v 1.29 2006/07/19 22:25:06 tews Exp $
  * 
  *)
 
 
-value ocamlp4_version = "0.2";
+value ocamlp4_version = "0.3";
 
 (****************************************************************************
  *      BEGIN RELOC SECTION
@@ -79,7 +79,7 @@ struct
 	| Loc
 	| Antiquot of MLast.loc and string ]
   ;
-    value loc =
+    value _loc =
 	let nowhere =
 	  {(Lexing.dummy_pos) with Lexing.pos_lnum = 1; Lexing.pos_cnum = 0 } in
 	  (nowhere,nowhere);
@@ -188,6 +188,7 @@ value class_sig_item = Grammar.Entry.create gram "class signature item";
 value class_str_item = Grammar.Entry.create gram "class structure item";
 value let_binding = Grammar.Entry.create gram "let_binding";
 value type_declaration = Grammar.Entry.create gram "type_declaration";
+value type_kind = Grammar.Entry.create gram "type_kind";
 value with_constr = Grammar.Entry.create gram "with_constr";
 value a_seq = Grammar.Entry.create gram "a_seq";
 value a_tup = Grammar.Entry.create gram "a_tup";
@@ -349,13 +350,17 @@ value rec mklistpat = fun
  * rewrites a right associative expr_ident like A.(B.(C.d))
  * into a left associative form, ie ((A.B).C).d
  *)
-
-value mkexprident loc i j =
-  loop (Qast.Node "ExUid" [Qast.Loc; i]) j where rec loop m =
-    fun
-    [ Qast.Node "ExAcc" [_; x; y] ->
-        loop (Qast.Node "ExAcc" [Qast.Loc; m; x]) y
-    | e -> Qast.Node "ExAcc" [Qast.Loc; m; e] ]
+value mkexprident _loc ids = match ids with
+  [ [] -> Stdpp.raise_with_loc _loc (Stream.Error "illegal long identifier")
+  | [ id :: ids ] ->
+      let rec loop m = fun
+        [ [ id :: ids ] -> loop 
+	    (* <:expr< $m$ . $id$ >>  *)
+	    (Qast.Node "ExAcc" [Qast.Loc; m; id])
+	    ids
+        | [] -> m ]
+      in
+	loop id ids ]
 ;
 
 
@@ -787,19 +792,19 @@ value rec is_expr_constr_call  =
 
 (* pa_o loc 307 *)
 
-value rec constr_expr_arity loc = 
+value rec constr_expr_arity _loc = 
   fun
   [ (* <:expr< $uid:c$ >> *) 
     Qast.Node "ExUid" [_; Qast.Str c] ->
       try List.assoc c constr_arity.val with [ Not_found -> 0 ]
   
   | (* <:expr< $uid:_$.$e$ >> *) 
-    Qast.Node "ExAcc" [_; Qast.Node "ExUid" _; e] -> constr_expr_arity loc e
+    Qast.Node "ExAcc" [_; Qast.Node "ExUid" _; e] -> constr_expr_arity _loc e
   
   | (* <:expr< $e$ $_$ >> *) 
     Qast.Node "ExApp" [_; e; _] ->
       if is_expr_constr_call e then
-        Stdpp.raise_with_loc loc (Stream.Error "currified constructor")
+        Stdpp.raise_with_loc _loc (Stream.Error "currified constructor")
       else 1
   | _ -> 1 ]
 ;
@@ -826,19 +831,19 @@ value rec is_patt_constr_call =
 
 (* pa_o loc 327 *)
 
-value rec constr_patt_arity loc =
+value rec constr_patt_arity _loc =
   fun
   [ (* <:patt< $uid:c$ >> *) 
     Qast.Node "PaUid" [_; Qast.Str c] ->
       try List.assoc c constr_arity.val with [ Not_found -> 0 ]
 
   | (* <:patt< $uid:_$.$p$ >> *)
-    Qast.Node "PaAcc" [_; Qast.Node "ExUid" _; p] -> constr_patt_arity loc p
+    Qast.Node "PaAcc" [_; Qast.Node "ExUid" _; p] -> constr_patt_arity _loc p
 
   | (* <:patt< $p$ $_$ >> *)
     Qast.Node "PaApp" [_; p; _ ] ->
       if is_patt_constr_call p then
-        Stdpp.raise_with_loc loc (Stream.Error "currified constructor")
+        Stdpp.raise_with_loc _loc (Stream.Error "currified constructor")
       else 1
   | _ -> 1 ]
 ;
@@ -903,8 +908,8 @@ value choose_tvar qast_tpl =
  * recognize the function syntactic suguar in let bindings, 
  * ie. recognize id <patt> ... which is an abbreviation 
  * of id = fun <patt> ...  
- * If recognized, patt_lid returns Some(loc_of_id, id, patt_list) 
- * otherwise none
+ * If recognized, patt_lid returns Some(id, patt_list) 
+ * otherwise None
  * 
  * pa_o loc 361
  *)
@@ -982,7 +987,7 @@ value ba_acc_term array_mod acc arr args =
 type bigarray_args = 
     [Arrayn of list Qast.t | Genarray of Qast.t | UseGenarray];
 
-value bigarray_get loc arr arg =
+value bigarray_get _loc arr arg =
   let rec qast_head n = fun
     [ Qast.Cons hd tl -> 
 	if n = 0 then UseGenarray 
@@ -1036,7 +1041,7 @@ value bigarray_get loc arr arg =
  * 
  * pa_o loc 386
  *)
-value bigarray_set loc var newval =
+value bigarray_set _loc var newval =
   match var with
   [ (* <:expr< Bigarray.Array1.get $arr$ $c1$ >> *) 
     Qast.Node "ExApp"
@@ -1128,7 +1133,7 @@ value bigarray_set loc var newval =
 
 EXTEND
   GLOBAL: ctyp poly_type str_item expr patt let_binding type_declaration
-    module_expr module_type sig_item with_constr class_type
+    module_expr module_type sig_item with_constr class_type type_kind
     class_expr class_str_item class_sig_item;
   
   module_expr:
@@ -1336,8 +1341,8 @@ EXTEND
 	  Qast.Node "ExFor" [Qast.Loc; i; e1; e2; df; get_seq e]
       | "while"; e1 = SELF; "do"; e2 = SELF; "done" ->
 	  Qast.Node "ExWhi" [Qast.Loc; e1; get_seq e2]
-	      
-
+      | "object"; cspo = SOPT class_self_patt; cf = class_structure; "end" ->
+          Qast.Node "ExObj" [Qast.Loc; cspo; cf ]
       ]
     | "tuple"
 	[ e = SELF; ","; el = SLIST1 (expr LEVEL ":=") SEP "," ->
@@ -1352,7 +1357,7 @@ EXTEND
 	       [Qast.Loc; e1; Qast.Node "ExLid" [Qast.Loc; Qast.Str "val"]];
 	     e2]
       | e1 = SELF; "<-"; e2 = expr LEVEL "expr1" ->
-          match bigarray_set loc e1 e2 with
+          match bigarray_set _loc e1 e2 with
               [ Some e -> e
               | None -> Qast.Node "ExAss" [Qast.Loc; e1; e2]
 	      ] ]
@@ -1575,7 +1580,7 @@ EXTEND
 
     | "apply" LEFTA
       [ e1 = SELF; e2 = SELF ->
-          match constr_expr_arity loc e1 with
+          match constr_expr_arity _loc e1 with
           [ 1 -> Qast.Node "ExApp" [ Qast.Loc; e1; e2]
           | _ -> 
               match e2 with
@@ -1592,7 +1597,7 @@ EXTEND
 	    Qast.Node "ExAre" [Qast.Loc; e1; e2]
 	| e1 = SELF; "."; "["; e2 = SELF; "]" -> 
 	    Qast.Node "ExSte" [Qast.Loc; e1; e2]
-	| e1 = SELF; "."; "{"; e2 = SELF; "}" -> bigarray_get loc e1 e2
+	| e1 = SELF; "."; "{"; e2 = SELF; "}" -> bigarray_get _loc e1 e2
      	| e1 = SELF; "."; e2 = SELF -> Qast.Node "ExAcc" [Qast.Loc; e1; e2]]
 
     | "~-" NONA
@@ -1619,7 +1624,7 @@ EXTEND
 	| s = a_CHAR -> Qast.Node "ExChr" [Qast.Loc; s]
 	| UIDENT "True" -> Qast.Node "ExUid" [Qast.Loc; Qast.Str " True"]
 	| UIDENT "False" -> Qast.Node "ExUid" [Qast.Loc; Qast.Str " False"]
-        | i = expr_ident -> i
+	| ids = expr_ident -> mkexprident _loc ids
 	| s = "false" -> Qast.Node "ExUid" [Qast.Loc; Qast.Str "False"]
 	| s = "true" -> Qast.Node "ExUid" [Qast.Loc; Qast.Str "True"]
         | "["; "]" -> Qast.Node "ExUid" [Qast.Loc; Qast.Str "[]"]
@@ -1690,9 +1695,10 @@ EXTEND
 
   expr_ident:
     [ RIGHTA
-	[ i = a_LIDENT -> Qast.Node "ExLid" [Qast.Loc; i]
-	| i = a_UIDENT -> Qast.Node "ExUid" [Qast.Loc; i]
-	| i = a_UIDENT; "."; j = SELF -> mkexprident Qast.Loc i j ] ]
+      [ i = a_LIDENT -> [ Qast.Node "ExLid" [Qast.Loc; i] ]
+      | i = a_UIDENT -> [ Qast.Node "ExUid" [Qast.Loc; i] ]
+      | i = a_UIDENT; "."; j = SELF -> 
+	  [(Qast.Node "ExUid" [Qast.Loc; i]) :: j ] ] ]
   ;
 
   fun_def:
@@ -1736,7 +1742,7 @@ EXTEND
 
     | LEFTA
       [ p1 = SELF; p2 = SELF ->
-          match constr_patt_arity loc p1 with
+          match constr_patt_arity _loc p1 with
           [ 1 -> Qast.Node "PaApp" [Qast.Loc; p1; p2]
           | n ->
               let p2 =
@@ -1852,34 +1858,25 @@ EXTEND
   ;
 
   type_kind:
-    [ [ "private"; "{"; ldl = SLIST1 label_declaration SEP ";"; "}" ->
-          Qast.Node "TyRec" [Qast.Loc; Qast.Bool True; ldl]
-      | "private"; OPT "|"; cdl = SLIST1 constructor_declaration SEP "|" -> 
-	  Qast.Node "TySum" [Qast.Loc; Qast.Bool True; cdl]
+    [ [ "private"; tk = type_kind -> Qast.Node "TyPrv" [Qast.Loc; tk]
       | test_constr_decl; 
 	OPT "|"; cdl = SLIST1 constructor_declaration SEP "|" -> 
-	  Qast.Node "TySum" [Qast.Loc; Qast.Bool False; cdl]
+	  Qast.Node "TySum" [Qast.Loc; cdl]
       | t = ctyp -> t
-      | t = ctyp; "="; "private"; 
-	  "{"; ldl = SLIST1 label_declaration SEP ";"; "}" ->
-	    Qast.Node "TyMan" 
-	      [Qast.Loc; t;
-	       Qast.Node "TyRec" [Qast.Loc; Qast.Bool True; ldl]]
+      | t = ctyp; "="; "private"; tk = type_kind ->
+      	  Qast.Node "TyMan" 
+      	    [Qast.Loc; t;
+	     Qast.Node "TyPrv" [Qast.Loc; tk]]
       | t = ctyp; "="; "{"; ldl = SLIST1 label_declaration SEP ";"; "}" ->
 	    Qast.Node "TyMan" 
 	      [Qast.Loc; t;
-	       Qast.Node "TyRec" [Qast.Loc; Qast.Bool False; ldl]]
-      | t = ctyp; "="; "private"; 
-	  OPT "|"; cdl = SLIST1 constructor_declaration SEP "|" ->
-	    Qast.Node "TyMan" 
-	      [Qast.Loc; t;
-	       Qast.Node "TySum" [Qast.Loc; Qast.Bool True; cdl]]
+	       Qast.Node "TyRec" [Qast.Loc; ldl]]
       | t = ctyp; "="; OPT "|"; cdl = SLIST1 constructor_declaration SEP "|" ->
 	    Qast.Node "TyMan" 
 	      [Qast.Loc; t;
-	       Qast.Node "TySum" [Qast.Loc; Qast.Bool False; cdl]]
+	       Qast.Node "TySum" [Qast.Loc; cdl]]
       | "{"; ldl = SLIST1 label_declaration SEP ";"; "}" ->
-	  Qast.Node "TyRec" [Qast.Loc; Qast.Bool False; ldl]
+	  Qast.Node "TyRec" [Qast.Loc; ldl]
       ] ]
   ;
 
@@ -1905,6 +1902,7 @@ EXTEND
      * the following ctyp grammar entry (but still much less than 
      * the ctyp of the revised syntax).
      *)
+
   ctyp:
     [ [ t1 = SELF; "as"; "'"; i = ident -> 
 	  Qast.Node "TyAli" 
@@ -2288,21 +2286,22 @@ EXTEND
  ****************************************************************************)
 
 
+  (* give a singleton list back; it will be converted by mkexprident *)
   expr_ident:
-    [ [ a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT -> [ antiquot "" _loc a ] ] ]
   ;
 
   patt_label_ident: LEVEL "simple"
-    [ [ a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
 
 
   type_patt:
-    [ [ a = ANTIQUOT "locstr" -> antiquot "locstr" loc a ] ]
+    [ [ a = ANTIQUOT "locstr" -> antiquot "locstr" _loc a ] ]
   ;
 
   mod_ident:
-    [ [ a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
 
   clty_longident:
@@ -2314,16 +2313,16 @@ EXTEND
   ;
 
   direction_flag:
-    [ [ a = ANTIQUOT "to" -> antiquot "to" loc a ] ]
+    [ [ a = ANTIQUOT "to" -> antiquot "to" _loc a ] ]
   ;
 
   class_signature:
-    [ [ a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
 
   meth_list:
     [ [ a = ANTIQUOT "list"; v = SOPT [ ";"; ".." -> Qast.Bool True ] -> 
-	  (antiquot "list" loc a, o2b v)
+	  (antiquot "list" _loc a, o2b v)
       ] ]
   ;
 
@@ -2342,23 +2341,23 @@ END;
 
 EXTEND
   module_expr: LEVEL "simple"
-    [ [ a = ANTIQUOT "mexp" -> antiquot "mexp" loc a
-      | a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT "mexp" -> antiquot "mexp" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
   
   str_item: LEVEL "top"
-    [ [ a = ANTIQUOT "stri" -> antiquot "stri" loc a
-      | a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT "stri" -> antiquot "stri" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
   
 
   module_type: LEVEL "simple"
-    [ [ a = ANTIQUOT "mtyp" -> antiquot "mtyp" loc a
-      | a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT "mtyp" -> antiquot "mtyp" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
   sig_item: LEVEL "top"
-    [ [ a = ANTIQUOT "sigi" -> antiquot "sigi" loc a
-      | a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT "sigi" -> antiquot "sigi" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
   expr: LEVEL "top"
     [ [ 
@@ -2379,9 +2378,9 @@ EXTEND
   expr: LEVEL "simple"
     [ [ 
 			(* the exp antiquotation in other nodes, eg str_item *)
-	a = ANTIQUOT "exp" -> antiquot "exp" loc a
+	a = ANTIQUOT "exp" -> antiquot "exp" _loc a
 				(* the expr antiquotation inside expressions *)
-      | a = ANTIQUOT -> antiquot "" loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
 
 
       ] ]
@@ -2396,8 +2395,8 @@ EXTEND
       ] ];
 
   patt: LEVEL "simple"
-    [ [ a = ANTIQUOT "pat" -> antiquot "pat" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "pat" -> antiquot "pat" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
 
 
     ] ]
@@ -2413,94 +2412,94 @@ EXTEND
   ;
 
   ctyp: LEVEL "simple"
-    [ [ a = ANTIQUOT "typ" -> antiquot "typ" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "typ" -> antiquot "typ" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       (* | "("; tl = a_list; ")" -> Qast.Node "TyTup" [Qast.Loc; tl]  *)
       ] ]
   ;
 
   class_expr: LEVEL "simple"
-    [ [ a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
 
   class_str_item:
-    [ [ a = ANTIQUOT "cli" -> antiquot "cli" loc a
-      | a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT "cli" -> antiquot "cli" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
 
   class_sig_item:
-    [ [ a = ANTIQUOT "clsi" -> antiquot "clsi" loc a 
-      | a = ANTIQUOT -> antiquot "" loc a ] ]
+    [ [ a = ANTIQUOT "clsi" -> antiquot "clsi" _loc a 
+      | a = ANTIQUOT -> antiquot "" _loc a ] ]
   ;
 
   (* put the antiquotation into class_signature, which is part of 
    * class_type (see local antiquotations above)
    * 
    * class_type:
-   *   [ [ a = ANTIQUOT -> antiquot "" loc a ] ]
+   *   [ [ a = ANTIQUOT -> antiquot "" _loc a ] ]
    * ;
    *)
 
 
   a_list:
-    [ [ a = ANTIQUOT "list" -> antiquot "list" loc a ] ]
+    [ [ a = ANTIQUOT "list" -> antiquot "list" _loc a ] ]
   ;
 
   a_tup:
-    [ [ a = ANTIQUOT "tup" -> antiquot "tup" loc a ] ]
+    [ [ a = ANTIQUOT "tup" -> antiquot "tup" _loc a ] ]
   ;
 
   a_seq:
-    [ [ a = ANTIQUOT "seq" -> antiquot "seq" loc a ] ]
+    [ [ a = ANTIQUOT "seq" -> antiquot "seq" _loc a ] ]
   ;
 
   a_opt:
-    [ [ a = ANTIQUOT "opt" -> antiquot "opt" loc a ] ]
+    [ [ a = ANTIQUOT "opt" -> antiquot "opt" _loc a ] ]
   ;
 
   a_UIDENT:
-    [ [ a = ANTIQUOT "uid" -> antiquot "uid" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "uid" -> antiquot "uid" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       | i = UIDENT -> Qast.Str i ] ]
   ;
   a_LIDENT:
-    [ [ a = ANTIQUOT "lid" -> antiquot "lid" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "lid" -> antiquot "lid" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       | i = LIDENT -> Qast.Str i ] ]
   ;
   a_INT:
-    [ [ a = ANTIQUOT "int" -> antiquot "int" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "int" -> antiquot "int" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       | s = INT -> Qast.Str s ] ]
   ;
   a_INT32:
-    [ [ a = ANTIQUOT "int32" -> antiquot "int32" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "int32" -> antiquot "int32" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       | s = INT32 -> Qast.Str s ] ]
   ;
   a_INT64:
-    [ [ a = ANTIQUOT "int64" -> antiquot "int64" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "int64" -> antiquot "int64" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       | s = INT64 -> Qast.Str s ] ]
   ;
   a_NATIVEINT:
-    [ [ a = ANTIQUOT "nativeint" -> antiquot "nativeint" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "nativeint" -> antiquot "nativeint" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       | s = NATIVEINT -> Qast.Str s ] ]
   ;
   a_FLOAT:
-    [ [ a = ANTIQUOT "flo" -> antiquot "flo" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "flo" -> antiquot "flo" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       | s = FLOAT -> Qast.Str s ] ]
   ;
   a_STRING:
-    [ [ a = ANTIQUOT "str" -> antiquot "str" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "str" -> antiquot "str" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       | s = STRING -> Qast.Str s ] ]
   ;
   a_CHAR:
-    [ [ a = ANTIQUOT "chr" -> antiquot "chr" loc a
-      | a = ANTIQUOT -> antiquot "" loc a
+    [ [ a = ANTIQUOT "chr" -> antiquot "chr" _loc a
+      | a = ANTIQUOT -> antiquot "" _loc a
       | s = CHAR -> Qast.Str s ] ]
   ;
 END;
@@ -2547,9 +2546,14 @@ do {
 (* 
  * In the revised syntax ctyp covers quantified types and constructions 
  * that appear only in type declarations. In the original syntax ctyp 
- * is much smaller. We therefore create a new quotation for the
- * type_declaration entry. The new quotation is actually called typedef, 
- * because that's the name of the metasymbol from the Ocaml documentation.
+ * is much smaller. We therefore create two new quotations. 
+ * First type_info for those parts of ctyp that really belong to type 
+ * declarations (and not to type expressions). Second typedef for type 
+ * declarations as a whole. Quantified types are included in the ctyp 
+ * quotation, therefore ctyp is renamed to ptyp. This gives two 
+ * quotation expanders (ptyp, type_info) for the same type (MLast.ctyp),
+ * which is a bit odd. However, the type_info quotations simplifies 
+ * paqo_o a lot.
  *
  * Further we include the quantified types in the ctyp quotation. 
  *)
@@ -2563,6 +2567,16 @@ do {
   Quotation.add "typedef" (apply_entry typedef_eoi)
 };
 
+let type_info_eoi = Grammar.Entry.create gram "type information" in
+do {
+  EXTEND
+   type_info_eoi:
+      [ [ x = type_kind; EOI -> x ] ]
+    ;
+  END;
+  Quotation.add "type_info" (apply_entry type_info_eoi)
+};
+
 let poly_type_eoi = Grammar.Entry.create gram "type" in
 do {
   EXTEND
